/*
 * Segment 8 Parametrierung und Konfiguration, SWE 8.2 Parametrierung - Bibliothek
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.puk.param.lib;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.Map.Entry;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.OneSubscriptionPerSendData;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.puk.param.lib.ParameterInfo.SerializableParameterInfo;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Clientseitige Implementierung der Funktionen für den Zugriff auf die Daten
 * und die Objekthierarchie der Parametrierung.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class ParameterManager implements Hierarchie, ClientReceiverInterface, ClientSenderInterface {

	/** wiederverwendbares leeres Systemobjekt-Array. */
	private static final SystemObject[] EMPTY_SYSOBJ_ARRAY = new SystemObject[0];

	/**
	 * Interner Thread zur Entkopplung der Weitergabe von Ergebnissen der
	 * Anfragen an die Parametrierung von weiteren Datenverteilerzugriffen..
	 *
	 * @author BitCtrl Systems GmbH, Peuker
	 */
	private static final class ErgebnisAuswerter extends Thread {

		/**
		 * Logger für Meldungsausgabe.
		 */
		private final Debug logger = Debug.getLogger();

		/**
		 * Liste der weiterzureichenden Ergebnisdaten.
		 */
		private final LinkedList<Data> ergebnisDaten = new LinkedList<>();

		/**
		 * der Parametermanager, dessen Antworten weitergereicht werden sollen.
		 */
		private final ParameterManager manager;

		/**
		 * Konstruktor.
		 *
		 * @param manager
		 *            der aufrufende Parametermanager
		 */
		ErgebnisAuswerter(final ParameterManager manager) {
			super("ParameterErgebnisAuswerter");
			this.manager = manager;
		}

		/**
		 * fügt einen weiterzureichenden Datensatz hinzu.
		 *
		 * @param data
		 *            die Daten
		 */
		void add(final Data data) {
			synchronized (ergebnisDaten) {
				ergebnisDaten.add(data);
			}
		}

		/**
		 * {@inheritDoc}.<br>
		 *
		 * @see java.lang.Thread#run()
		 */
		@Override
		public void run() {
			do {
				logger.finer("Auswerten von Ergebnissen");
				synchronized (ergebnisDaten) {
					while (!ergebnisDaten.isEmpty()) {
						final Data data = ergebnisDaten.getFirst();
						ergebnisDaten.removeFirst();
						manager.behandleErgebnis(data);
					}
				}
				try {
					synchronized (this) {
						this.wait(MethodenBibliothek.MILLIS_PER_MINUTE);
					}
				} catch (final InterruptedException e) {
					logger.warning("ErgebnisAuswerter wurde unterbrochen: ", new Object[] { e });
				}
			} while (true);
		}
	}

	/**
	 * Logger für Meldungsausgabe.
	 */
	private static final Debug LOGGER = Debug.getLogger();

	/**
	 * der Thread zum Weiterleiten der Anfrage-Antworten.
	 */
	private final ErgebnisAuswerter ergebnisAuswerter = new ErgebnisAuswerter(this);

	/**
	 * die verwendete Datenverteilerverbindung.
	 */
	private final ClientDavInterface verbindung;

	/**
	 * die Datenbeschreibung für eine Anfrage an die Parametrierung.
	 */
	private final DataDescription anfrageDesc;

	/**
	 * die Datenbeschreibung für die Antwort der Parametrierung auf eine
	 * Anfrage.
	 */
	private final DataDescription antwortDesc;

	/**
	 * die versendeten Anfragen, die auf eine Antwort warten.
	 */
	private final HashMap<Long, ParametrierungsAnfrage> anfragen = new HashMap<>();

	/**
	 * markiert die gültigkeit des ParameterManagers.<br>
	 * Die Instanz wird ungültig, wenn die verwendete Verbindung geschlosssen
	 * wurde oder eine Anmeldung nicht erfolgreich war.
	 */
	private boolean gueltig = true;

	/**
	 * der Beobachter, der angemledete Clients über veränderte Parameter
	 * informiert.
	 */
	private final ParameterBeobachter beobachter;

	/**
	 * konstruiert eine Instanz der Klasse für die übergebene
	 * Datenverteilerverbindung.
	 *
	 * @param verbindung
	 *            die verwendete Datenverteilerverbindung
	 */
	ParameterManager(final ClientDavInterface verbindung) {
		this.verbindung = verbindung;

		beobachter = new ParameterBeobachter(this.verbindung);
		ergebnisAuswerter.start();

		final DataModel model = this.verbindung.getDataModel();
		AttributeGroup atg = model.getAttributeGroup("atg.parametrierungsAnfrage");
		if (null == atg) {
			atg = model.getAttributeGroup("atg.parametrierungSchnittstelle");
		}
		anfrageDesc = new DataDescription(atg, model.getAspect("asp.anfrage"), (short) 0);
		atg = model.getAttributeGroup("atg.parametrierungsAntwort");
		if (null == atg) {
			atg = model.getAttributeGroup("atg.parametrierungSchnittstelle");
		}
		antwortDesc = new DataDescription(atg, model.getAspect("asp.antwort"), (short) 0);

		this.verbindung.subscribeReceiver(this, new SystemObject[] { this.verbindung.getLocalApplicationObject() },
				antwortDesc, ReceiveOptions.normal(), ReceiverRole.drain());

		try {
			this.verbindung.subscribeSender(this, this.verbindung.getLocalConfigurationAuthority(), anfrageDesc,
					SenderRole.sender());
		} catch (final OneSubscriptionPerSendData e) {
			throw new RuntimeException(
					"Es besteht bereits eine Anmeldung für die Benutzung der Parameter-Schnittstelle für die Applikation !",
					e);
		}
	}

	/**
	 * installiert den übergebenen Listener als {@link ParameterListener} für
	 * die übergebenen Parameterspezifikationen.
	 *
	 * @param listener
	 *            der Listener
	 * @param parameter
	 *            dte Spezifikationen
	 * @throws ParameterClientException
	 *             die übergebenen Parameterspezifikationen enthielten ungültige
	 *             Einträge.
	 */
	public void addListener(final ParameterListener listener, final ParameterInfo... parameter)
			throws ParameterClientException {
		beobachter.addListener(listener, parameter);
	}

	/**
	 * ordnet einen Ergebnisdatensatz der entsprechenden Anfrage zu.
	 *
	 * @param antwortDaten
	 *            der empfangene Antwortdatensatz
	 */
	void behandleErgebnis(final Data antwortDaten) {
		if (null != antwortDaten) {
			final long id = antwortDaten.getUnscaledValue("AnfrageIndex").longValue();
			final ParametrierungsAnfrage anfrage = anfragen.get(id);
			if (anfrage != null) {
				anfrage.setErgebnis(antwortDaten);
				anfragen.remove(id);
			}
		}
	}

	/** {@inheritDoc} */
	@Override
	public void dataRequest(final SystemObject object, final DataDescription dataDescription, final byte state) {
		ParameterManager.LOGGER.warning("Unerwartete Meldung der Sendesteuerung",
				new Object[] { object, dataDescription, state });
	}

	/** {@inheritDoc} */
	@Override
	public Collection<SystemObject> getAenderbareObjekte(final ParameterInfo parameter) {

		final ParametrierungsAnfrage anfrage = new ParametrierungsAnfrage(
				ParametrierungsAnfrageTyp.SUCHE_AENDERBARE_OBJEKTE, parameter.getSerializableParameterInfo(verbindung));
		anfragen.put(anfrage.getId(), anfrage);

		Collection<SystemObject> result = null;
		try {
			final Long[] ids = (Long[]) anfrage.warteAufErgebnis(verbindung, anfrageDesc);
			result = new ArrayList<>();
			for (final long id : ids) {
				result.add(verbindung.getDataModel().getObject(id));
			}
		} catch (final ParameterClientException e) {
			Debug.getLogger().fine("Fehler beim lesen änderbarer Objekte", e);
		}
		return result;
	}

	/**
	 * ermittelt die Parameter für die Menge der übergebenen
	 * Parameterspezifikationen ohne die Parameterdaten, wenn die Parameter
	 * nicht abgefragt werden konnten, wird {@code null} zurückgeliefert.
	 *
	 * @param info
	 *            die Spezifikationen
	 * @return die ermittelten Parameter oder {@code null}
	 * @throws ParameterClientException
	 *             die Parameter konnten nicht ermittelt werden
	 */
	public Parameter[] getParameterOhneDaten(final ParameterInfo... info) throws ParameterClientException {
		return getParameter(false, info);
	}

	/**
	 * ermittelt die Parameter für die Menge der übergebenen
	 * Parameterspezifikationen inklusive der Parameterdaten, wenn die Parameter
	 * nicht abgefragt werden konnten, wird {@code null} zurückgeliefert.
	 *
	 * @param info
	 *            die Spezifikationen
	 * @return die ermittelten Parameter oder {@code null}
	 * @throws ParameterClientException
	 *             die Parameter konnten nicht ermittelt werden
	 */
	public Parameter[] getParameter(final ParameterInfo... info) throws ParameterClientException {
		return getParameter(true, info);
	}

	/**
	 * ermittelt die Parameter für die Menge der übergebenen
	 * Parameterspezifikationen, wenn die Parameter nicht abgefragt werden
	 * konnten, wird {@code null} zurückgeliefert.
	 *
	 * @param datenAbrufen
	 *            legt fest, ob Daten mit abgerufen werden sollen oder nicht
	 * @param info
	 *            die Spezifikationen
	 * @return die ermittelten Parameter oder {@code null}
	 * @throws ParameterClientException
	 *             die Parameter konnten nicht ermittelt werden
	 */
	public Parameter[] getParameter(final boolean datenAbrufen, final ParameterInfo... info)
			throws ParameterClientException {

		final SerializableParameterInfo[] anfrageParameter = new SerializableParameterInfo[info.length];
		for (int idx = 0; idx < info.length; idx++) {
			anfrageParameter[idx] = info[idx].getSerializableParameterInfo(verbindung);
		}

		final ParameterAbfrageObjekt anfrageObjekt = new ParameterAbfrageObjekt(anfrageParameter, datenAbrufen);

		final ParametrierungsAnfrage anfrage = new ParametrierungsAnfrage(ParametrierungsAnfrageTyp.PARAMETER_LESEN,
				anfrageObjekt);

		anfragen.put(anfrage.getId(), anfrage);

		final SerializableParameter[] antworten = (SerializableParameter[]) anfrage.warteAufErgebnis(verbindung,
				anfrageDesc);
		final Parameter[] parameter = new Parameter[info.length];
		for (int idx = 0; idx < info.length; idx++) {
			if (null == antworten[idx]) {
				parameter[idx] = null;
			} else {
				parameter[idx] = antworten[idx].erzeugeParameter(verbindung);
			}
		}

		return parameter;
	}

	/**
	 * {@inheritDoc}.
	 *
	 * @see de.bsvrz.puk.param.lib.Hierarchie#getParameterWurzel(java.util.Collection)
	 */
	@Override
	public Map<ParameterInfo, SystemObject> getParameterWurzel(final Collection<ParameterInfo> parameter) {
		final ParameterInfo[] infoFeld = parameter.toArray(new ParameterInfo[parameter.size()]);
		final SystemObject[] ergebnisFeld = getParameterWurzel(infoFeld);
		final Map<ParameterInfo, SystemObject> ergebnis = new HashMap<>();

		for (int idx = 0; idx < infoFeld.length; idx++) {
			ergebnis.put(infoFeld[idx], ergebnisFeld[idx]);
		}

		return ergebnis;
	}

	/**
	 * {@inheritDoc}.
	 *
	 * @see de.bsvrz.puk.param.lib.Hierarchie#getParameterWurzel(de.bsvrz.puk.param.lib.ParameterInfo)
	 */
	@Override
	public SystemObject getParameterWurzel(final ParameterInfo parameter) {
		return getParameterWurzel(new ParameterInfo[] { parameter })[0];
	}

	/**
	 * {@inheritDoc}.
	 *
	 * @see de.bsvrz.puk.param.lib.Hierarchie#getParameterWurzel(de.bsvrz.puk.param.lib.ParameterInfo[])
	 */
	@Override
	public SystemObject[] getParameterWurzel(final ParameterInfo[] parameter) {

		final SerializableParameterInfo[] anfrageParameter = new SerializableParameterInfo[parameter.length];
		for (int idx = 0; idx < parameter.length; idx++) {
			anfrageParameter[idx] = parameter[idx].getSerializableParameterInfo(verbindung);
		}

		final ParametrierungsAnfrage anfrage = new ParametrierungsAnfrage(
				ParametrierungsAnfrageTyp.SUCHE_PARAMETER_WURZEL, anfrageParameter);
		anfragen.put(anfrage.getId(), anfrage);
		SystemObject[] result;
		try {
			final Long[] ids = (Long[]) anfrage.warteAufErgebnis(verbindung, anfrageDesc);
			final Collection<SystemObject> liste = new ArrayList<>();
			for (final long id : ids) {
				liste.add(verbindung.getDataModel().getObject(id));
			}
			result = liste.toArray(ParameterManager.EMPTY_SYSOBJ_ARRAY);
		} catch (final ParameterClientException e) {
			Debug.getLogger().fine("Fehler beim ermiiteln der Parameterwurzel", e);
			result = null;
		}
		return result;
	}

	/**
	 * ermittelt, ob der Parametermanger gültig ist, d.h. für Abfragen verwendet
	 * werden kann.
	 *
	 * @return die gültigkeit
	 */
	public final boolean isGueltig() {
		return gueltig;
	}

	/** {@inheritDoc} */
	@Override
	public boolean isRequestSupported(final SystemObject object, final DataDescription dataDescription) {
		return false;
	}

	/**
	 * entfernt den angegebenen {@link ParameterListener} als Listener für die
	 * angegebenen Parameterspezifikationen.
	 *
	 * @param listener
	 *            der Listener
	 * @param parameter
	 *            die Menge der abzumeldenden Spezifikationen
	 */
	public void removeListener(final ParameterListener listener, final ParameterInfo... parameter) {
		beobachter.removeListener(listener, parameter);
	}

	/**
	 * sendet die Daten der übergebenen Parameter, um sie in der Parametrierung
	 * zu setzen.
	 *
	 * @param parameter
	 *            die zu setzenden Parameter
	 * @throws ParameterClientException
	 *             die Parameter konnten nicht gesetzt werden
	 */
	public void setParameter(final Parameter... parameter) throws ParameterClientException {

		final SerializableParameter[] parameterListe = new SerializableParameter[parameter.length];
		for (int idx = 0; idx < parameter.length; idx++) {
			parameterListe[idx] = parameter[idx].getSerializableParameter(verbindung);
		}
		final ParametrierungsAnfrage anfrage = new ParametrierungsAnfrage(ParametrierungsAnfrageTyp.PARAMETER_SCHREIBEN,
				parameterListe);
		anfragen.put(anfrage.getId(), anfrage);
		anfrage.warteAufErgebnis(verbindung, anfrageDesc);
	}

	/**
	 * sendet die Daten der übergebenen Parameter, um sie in der Parametrierung
	 * zu setzen.
	 *
	 * Als Ergebnis wird eine Map mit Fehlern geliefert, wenn die Version der
	 * Parametrierung dies unterstützt. Wenn alle Parameter gesetzt werden
	 * konnten ist diese leer, ansonsten enthält Sie die Parameterinfos mit den
	 * entsprechenden Fehlertexten.
	 *
	 * @param parameter
	 *            die zu setzenden Parameter
	 * @return das Ergebnis der Operation
	 * @throws ParameterClientException
	 *             die Parameter konnten nicht gesetzt werden
	 */
	public Map<ParameterInfo, String> setParameterChecked(final Parameter... parameter)
			throws ParameterClientException {

		final Map<ParameterInfo, String> result = new HashMap<>();

		final SerializableParameter[] parameterListe = new SerializableParameter[parameter.length];
		for (int idx = 0; idx < parameter.length; idx++) {
			parameterListe[idx] = parameter[idx].getSerializableParameter(verbindung);
		}
		final ParametrierungsAnfrage anfrage = new ParametrierungsAnfrage(ParametrierungsAnfrageTyp.PARAMETER_SCHREIBEN,
				parameterListe);
		anfragen.put(anfrage.getId(), anfrage);
		final Serializable ergebnis = anfrage.warteAufErgebnis(verbindung, anfrageDesc);

		if (ergebnis instanceof HashMap<?, ?>) {
			for (final Entry<?, ?> entry : ((HashMap<?, ?>) ergebnis).entrySet()) {
				final Object key = entry.getKey();
				final Object value = entry.getValue();

				if (key instanceof SerializableParameterInfo && value instanceof String) {
					result.put(((SerializableParameterInfo) key).erzeugeParameterInfo(verbindung), (String) value);
				} else {
					Debug.getLogger().warning(
							"Die Parametrierung liefert ein unerwartetes Ergebnis: " + key + "/" + value + "!");
				}
			}
		} else {
			Debug.getLogger().warning(
					"Die Parametrierung liefert ein unerwartetes Ergebnis! Bitte die aktuelle Version verwenden!");
		}

		return result;
	}

	/**
	 * markiert den Parametermanger als ungültig.
	 */
	void ungueltigSetzen() {
		gueltig = false;
	}

	/** {@inheritDoc} */
	@Override
	public void update(final ResultData[] results) {
		for (final ResultData result : results) {
			ParameterManager.LOGGER.fine("UPDATE ParameterManager: " + result);
			if (result.getDataDescription().getAttributeGroup().equals(antwortDesc.getAttributeGroup())) {
				if (result.getDataDescription().getAspect().equals(antwortDesc.getAspect())) {
					ergebnisAuswerter.add(result.getData());
				}
			}
		}
		synchronized (ergebnisAuswerter) {
			ergebnisAuswerter.notifyAll();
		}
	}

	/**
	 * versendet einen Auftrag zur Erstellung eines Backups der
	 * Parameterdatenbank.
	 *
	 * @return das Ergebnis der Backuperstellung
	 */
	public Boolean requestBackup() {

		final ParametrierungsAnfrage anfrage = new ParametrierungsAnfrage(-1);
		anfragen.put(anfrage.getId(), anfrage);

		Boolean result = false;
		try {
			result = (Boolean) anfrage.warteAufErgebnis(verbindung, anfrageDesc);
		} catch (final ParameterClientException e) {
			Debug.getLogger().error(e.getLocalizedMessage(), e);
		}
		return result;
	}
}
